# noqa: D100
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Generator
    from pathlib import Path


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class Command:
    """Base class for all commands."""


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class Message(Command):
    """Message displayed to the user."""

    value: str


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class Call(Command):
    """Call another command; changes made by the called script are persisted."""

    command_line: str
    exit_on_error: bool = field(default=True, kw_only=True)
    exit_via_return_statement: bool = field(default=False, kw_only=True)


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class Execute(Command):
    """Call another executable; changes made by the called executable are not persisted."""

    command_line: str
    exit_on_error: bool = field(default=True, kw_only=True)
    exit_via_return_statement: bool = field(default=False, kw_only=True)

    # ----------------------------------------------------------------------
    def __post_init__(self) -> None:
        assert not self.exit_via_return_statement or self.exit_on_error


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class Set(Command):
    """Set an environment variable, overwriting any existing value."""

    name: str
    value_or_values: str | list[str] | None

    # ----------------------------------------------------------------------
    def EnumValues(self) -> Generator[str]:
        """Enumerate the values provided."""

        if isinstance(self.value_or_values, str):
            yield self.value_or_values
            return

        if isinstance(self.value_or_values, list):
            yield from self.value_or_values
            return

        if self.value_or_values is None:
            return

        assert False, type(self.value_or_values)  # noqa: B011, PT015 # pragma: no cover


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class Augment(Command):
    """Augment an environment variable."""

    name: str
    value_or_values: str | list[str]
    append_values: bool = field(default=False, kw_only=True)

    # ----------------------------------------------------------------------
    def EnumValues(self) -> Generator[str]:
        """Enumerate the values provided."""

        if isinstance(self.value_or_values, str):
            yield self.value_or_values
            return

        if isinstance(self.value_or_values, list):
            yield from self.value_or_values
            return

        assert False, type(self.value_or_values)  # noqa: B011, PT015 # pragma: no cover


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class Exit(Command):
    """Exit the script with an optional return code."""

    pause_on_success: bool = field(default=False, kw_only=True)
    pause_on_error: bool = field(default=False, kw_only=True)
    return_code: int | None = field(default=None, kw_only=True)


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class ExitOnError(Command):
    """Exit the script if an error was generated when executing the previous command."""

    variable_name: str | None = field(default=None)
    return_code: int | None = field(default=None)
    use_return_statement: bool = field(default=False, kw_only=True)

    # ----------------------------------------------------------------------
    def __post_init__(self) -> None:
        assert (
            (self.variable_name is None and self.return_code is None)
            or (self.variable_name is not None and self.return_code is None)
            or (self.variable_name is None and self.return_code is not None)
        ), (self.variable_name, self.return_code)


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class EchoOff(Command):
    """Disable printing of commands to the terminal."""


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class PersistError(Command):
    """Persist the current error value."""

    variable_name: str


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class PushDirectory(Command):
    """Push the current directory onto the directory stack and change to a new directory."""

    value: Path | None


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class PopDirectory(Command):
    """Pop a directory from the directory stack and change to that directory."""


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class Raw(Command):
    """Raw, shell-specific content that isn't altered during decoration."""

    value: str



# noqa: D100
import platform
import sys

from dataclasses import dataclass
from enum import auto, Enum
from functools import cached_property
from typing import Literal, TYPE_CHECKING

from dbrownell_Common.InflectEx import inflect
from semantic_version import Version as SemVer

if TYPE_CHECKING:
    from collections.abc import Generator
    from pathlib import Path

    from dbrownell_Common.Streams.DoneManager import DoneManager


# ----------------------------------------------------------------------
# |
# |  Public Types
# |
# ----------------------------------------------------------------------
class OperatingSystemType(Enum):
    """Specify the operating system."""

    Linux = auto()
    MacOS = auto()
    Windows = auto()

    # ----------------------------------------------------------------------
    @classmethod
    def Calculate(cls) -> OperatingSystemType:
        """Calculate the current operating system."""

        platform_str = sys.platform.lower()

        if platform_str.startswith("linux"):
            return OperatingSystemType.Linux

        if platform_str.startswith("darwin"):
            return OperatingSystemType.MacOS

        if platform_str.startswith("win32"):
            return OperatingSystemType.Windows

        msg = f"Unsupported platform: '{platform_str}'"
        raise Exception(msg)

    # ----------------------------------------------------------------------
    @cached_property
    def string_map(self) -> dict[str, OperatingSystemType | Literal["Generic"]]:
        """Return strings that represent operating systems."""

        results: dict[str, OperatingSystemType | Literal["Generic"]] = {
            e.name: e for e in OperatingSystemType
        }

        results["Generic"] = "Generic"

        return results


# ----------------------------------------------------------------------
class ArchitectureType(Enum):
    """Specify the system architecture."""

    x64 = auto()
    x86 = auto()
    ARM64 = auto()
    ARM = auto()

    # ----------------------------------------------------------------------
    @classmethod
    def Calculate(cls) -> ArchitectureType:
        """Calculate the current system architecture."""

        arch = platform.machine().lower()

        if arch == "arm64":
            return ArchitectureType.ARM64
        if arch == "amd":
            return ArchitectureType.ARM

        if arch not in ["x86_64", "amd64"]:
            msg = f"Unsupported architecture: '{arch}'"
            raise Exception(msg)

        return ArchitectureType.x64 if sys.maxsize > 2**32 else ArchitectureType.x86

    # ----------------------------------------------------------------------
    @cached_property
    def string_map(self) -> dict[str, ArchitectureType | Literal["Generic"]]:
        """Return strings that represent architectures."""

        results: dict[str, ArchitectureType | Literal["Generic"]] = {e.name: e for e in ArchitectureType}

        results["Generic"] = "Generic"

        return results


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class ToolInfo:
    """Information about a tool."""

    name: str
    """Name of the tool."""

    version: SemVer | None
    """Version of the tool, if a versioned directory was encountered."""

    operating_system: OperatingSystemType | Literal["Generic"] | None
    """Operating system of the tool, if an OS-specific directory was encountered."""

    architecture: ArchitectureType | Literal["Generic"] | None
    """Architecture of the tool, if an architecture-specific directory was encountered."""

    root_directory: Path
    """Root of the tool; all versioned directories are subdirectories of this."""

    versioned_directory: Path
    """Root of a specific version of the tool."""

    binary_directory: Path
    """Binary directory of a specific version of the tool."""

    # ----------------------------------------------------------------------
    def GeneratePotentialEnvFiles(self, file_extension: str = ".env") -> Generator[Path]:
        """Generate potential filenames for environment files, ordered from the least- to most-specific."""

        # ----------------------------------------------------------------------
        def ApplyStandardSuffixes(potential_suffixes: list[str]) -> None:
            assert not potential_suffixes
            potential_suffixes.append("")

        # ----------------------------------------------------------------------
        def ApplyVersionSuffixes(potential_suffixes: list[str]) -> None:
            if self.version is None:
                return

            potential_suffixes.append(f"-v{self.version}")

        # ----------------------------------------------------------------------
        def ApplyOperatingSystemSuffixes(potential_suffixes: list[str]) -> None:
            if self.operating_system is None:
                return

            suffix = (
                self.operating_system.name
                if isinstance(self.operating_system, OperatingSystemType)
                else self.operating_system
            )

            potential_suffixes.append(f"-{suffix}")
            potential_suffixes += [
                f"{potential_suffix}-{suffix}" for potential_suffix in potential_suffixes[1:-1]
            ]

        # ----------------------------------------------------------------------
        def ApplyArchitectureSuffixes(potential_suffixes: list[str]) -> None:
            if self.architecture is None:
                return

            suffix = (
                self.architecture.name
                if isinstance(self.architecture, ArchitectureType)
                else self.architecture
            )

            potential_suffixes.append(f"-{suffix}")
            potential_suffixes += [
                f"{potential_suffix}-{suffix}" for potential_suffix in potential_suffixes[1:-1]
            ]

        # ----------------------------------------------------------------------

        # Get a list of directories between the root and the versioned directory. Append "fake_file"
        # so that versioned_directory is included in the parents (without "fake_file", the implementation
        # thinks that versioned_directory is a file and not a directory).
        relative_paths = list(
            reversed((self.versioned_directory / "fake_file").relative_to(self.root_directory).parents)
        )

        # At most, there can be relative paths for:
        # 1) Root
        # 2) Version
        # 3) Operating System
        # 4) Architecture
        assert len(relative_paths) <= 4, relative_paths  # noqa: PLR2004

        # Root
        potential_suffixes: list[str] = []

        ApplyStandardSuffixes(potential_suffixes)
        ApplyVersionSuffixes(potential_suffixes)
        ApplyOperatingSystemSuffixes(potential_suffixes)
        ApplyArchitectureSuffixes(potential_suffixes)

        relative_path_offset = 0
        root = self.root_directory / relative_paths[relative_path_offset]

        yield from [root / f"{self.name}{suffix}{file_extension}" for suffix in potential_suffixes]

        # Version
        if self.version is not None:
            potential_suffixes = []

            ApplyStandardSuffixes(potential_suffixes)
            ApplyOperatingSystemSuffixes(potential_suffixes)
            ApplyArchitectureSuffixes(potential_suffixes)

            relative_path_offset += 1
            root = self.root_directory / relative_paths[relative_path_offset]

            yield from [root / f"{self.name}{suffix}{file_extension}" for suffix in potential_suffixes]

        # Operating System
        if self.operating_system is not None:
            potential_suffixes = []

            ApplyStandardSuffixes(potential_suffixes)
            ApplyArchitectureSuffixes(potential_suffixes)

            relative_path_offset += 1
            root = self.root_directory / relative_paths[relative_path_offset]

            yield from [root / f"{self.name}{suffix}{file_extension}" for suffix in potential_suffixes]

        # Architecture
        if self.architecture is not None:
            potential_suffixes = []

            ApplyStandardSuffixes(potential_suffixes)

            relative_path_offset += 1
            root = self.root_directory / relative_paths[relative_path_offset]

            yield root / f"{self.name}{file_extension}"


# ----------------------------------------------------------------------
# |
# |  Public Functions
# |
# ----------------------------------------------------------------------
def GetToolInfos(
    dm: DoneManager,
    tools_directory: Path,
    include_tools: set[str],
    exclude_tools: set[str],
    tool_versions: dict[str, SemVer],
    operating_system: OperatingSystemType,
    architecture: ArchitectureType,
    *,
    no_generic_operating_systems: bool = False,
    no_generic_architectures: bool = False,
) -> list[ToolInfo]:
    """Return information about tools in a specified directory."""

    results: list[ToolInfo] = []

    with dm.Nested(
        f"Parsing tools in '{tools_directory}'...",
        lambda: "{} found".format(inflect.no("tool", len(results))),
    ) as tool_dm:
        for tool_directory in tools_directory.iterdir():
            if not tool_directory.is_dir():
                continue

            root_directory = tool_directory
            tool_name = tool_directory.name

            if tool_name in exclude_tools:
                tool_dm.WriteVerbose(f"'{tool_name}' has been explicitly excluded.\n")
                continue

            if include_tools and tool_name not in include_tools:
                tool_dm.WriteVerbose(f"'{tool_name}' has not been explicitly included.\n")
                continue

            tool_directory, semantic_version = _ApplyVersionDir(  #  noqa: PLW2901
                tool_dm,
                tool_name,
                tool_directory,
                tool_versions,
            )

            tool_directory, operating_system_type = _ApplyOperatingSystemDir(  # noqa: PLW2901
                tool_dm,
                tool_name,
                tool_directory,
                operating_system,
                no_generic_operating_systems=no_generic_operating_systems,
            )

            tool_directory, architecture_type = _ApplyArchitectureDir(  # noqa: PLW2901
                tool_dm,
                tool_name,
                tool_directory,
                architecture,
                no_generic_architectures=no_generic_architectures,
            )

            bin_tool_directory = _ApplyBinDir(tool_dm, tool_directory)

            if tool_dm.result != 0:
                continue

            results.append(
                ToolInfo(
                    tool_name,
                    semantic_version,
                    operating_system_type,
                    architecture_type,
                    root_directory,
                    tool_directory,
                    bin_tool_directory,
                ),
            )

    return results


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
def _ApplyVersionDir(
    dm: DoneManager,
    tool_name: str,
    tool_directory: Path,
    tool_versions: dict[str, SemVer],
) -> tuple[Path, SemVer | None]:
    working_version_dirs: list[tuple[SemVer, Path]] = []

    for potential_dir in tool_directory.iterdir():
        if not potential_dir.is_dir():
            continue

        try:
            version = SemVer.coerce(potential_dir.name.removeprefix("v"))
            working_version_dirs.append((version, potential_dir))
        except ValueError:
            continue

    if working_version_dirs:
        working_version_dirs.sort(key=lambda x: x[0], reverse=True)

        working_versions = dict(working_version_dirs)

        desired_version = tool_versions.get(
            tool_name,
            next(iter(working_versions)),
        )

        updated_working_dir = working_versions.get(desired_version)
        if updated_working_dir is not None:
            return updated_working_dir, desired_version

        dm.WriteError(
            f"No directory found for version '{desired_version}' for the tool '{tool_name}' in '{tool_directory}'.\n"
        )

    return tool_directory, None


# ----------------------------------------------------------------------
def _ApplyOperatingSystemDir(
    dm: DoneManager,
    tool_name: str,
    tool_directory: Path,
    operating_system: OperatingSystemType,
    *,
    no_generic_operating_systems: bool,
) -> tuple[Path, OperatingSystemType | Literal["Generic"] | None]:
    operating_system_dirs: dict[str, tuple[Path, OperatingSystemType | Literal["Generic"]]] = {}

    for potential_dir in tool_directory.iterdir():
        if not potential_dir.is_dir():
            continue

        potential_mapped_info = operating_system.string_map.get(potential_dir.name)
        if potential_mapped_info is not None:
            operating_system_dirs[potential_dir.name] = (potential_dir, potential_mapped_info)

    if operating_system_dirs:
        if operating_system.name in operating_system_dirs:
            return operating_system_dirs[operating_system.name]

        if not no_generic_operating_systems and "Generic" in operating_system_dirs:
            return operating_system_dirs["Generic"]

        dm.WriteError(
            f"No directory found for '{operating_system.name}' for the tool '{tool_name}' in '{tool_directory}'.\n"
        )

    return tool_directory, None


# ----------------------------------------------------------------------
def _ApplyArchitectureDir(
    dm: DoneManager,
    tool_name: str,
    tool_directory: Path,
    architecture: ArchitectureType,
    *,
    no_generic_architectures: bool,
) -> tuple[Path, ArchitectureType | Literal["Generic"] | None]:
    architecture_dirs: dict[str, tuple[Path, ArchitectureType | Literal["Generic"]]] = {}

    for potential_dir in tool_directory.iterdir():
        if not potential_dir.is_dir():
            continue

        potential_mapped_info = architecture.string_map.get(potential_dir.name)
        if potential_mapped_info is not None:
            architecture_dirs[potential_dir.name] = (potential_dir, potential_mapped_info)

    if architecture_dirs:
        if architecture.name in architecture_dirs:
            return architecture_dirs[architecture.name]

        if not no_generic_architectures and "Generic" in architecture_dirs:
            return architecture_dirs["Generic"]

        dm.WriteError(
            f"No directory found for '{architecture.name}' for the tool '{tool_name}' in '{tool_directory}'.\n"
        )

    return tool_directory, None


# ----------------------------------------------------------------------
def _ApplyBinDir(
    dm: DoneManager,  # noqa: ARG001
    tool_directory: Path,
) -> Path:
    potential_dir = tool_directory / "bin"
    if potential_dir.is_dir():
        return potential_dir

    return tool_directory



# noqa: D100
import io
import logging
import os
import re

from typing import cast, TYPE_CHECKING

from dbrownell_Common.ContextlibEx import ExitStack
from dotenv import dotenv_values

from dbrownell_ToolsDirectory.Shell.Commands import Augment, Command, EchoOff, Raw, Set

if TYPE_CHECKING:
    from pathlib import Path

    from dbrownell_Common.Streams.DoneManager import DoneManager

    from dbrownell_ToolsDirectory.ToolInfo import ToolInfo


# ----------------------------------------------------------------------
def CreateShellCommands(
    dm: DoneManager,
    tool_infos: list[ToolInfo],
) -> list[Command]:
    r"""Create shell commands to set up the environment for the specified tools.

    For each tool, this function looks for one or more ``.env`` files (as
    determined by :meth:`ToolInfo.GeneratePotentialEnvFiles`) and loads
    environment variables from those files.

    When reading values from a ``.env`` file, any path segment that begins
    with ``./`` or ``.\`` is treated as relative to the directory that
    contains the ``.env`` file itself. The leading ``./`` or ``.\`` is
    replaced with the absolute path to the ``.env`` file's directory plus
    the appropriate path separator.

    Parent-directory references such as ``../`` are not resolved or
    rewritten by this logic; they are left in the value as-is. Only
    current-directory references using ``./`` or ``.\`` are expanded.

    Example:
    * Given an environment file ``/tools/Tool1/Tool1.env`` containing::
          RELATIVE_PATH=./bin
      the value will be expanded to::
          /tools/Tool1/bin

    """

    with dm.Nested("Creating shell commands...") as shell_dm:
        relative_path_re = re.compile(r"(?<!\.)(?P<current_dir>\.[/\\])")

        # ----------------------------------------------------------------------
        def RelativePathReplacement(_: re.Match, env_filename: Path) -> str:
            return str(env_filename.parent) + os.path.sep

        # ----------------------------------------------------------------------

        # Update dotenv's logger to suppress output
        logger_sink = io.StringIO()

        handler = logging.StreamHandler(logger_sink)

        with ExitStack(handler.close):
            logger = logging.getLogger("dotenv.main")

            logger.propagate = False
            logger.setLevel(logging.WARNING)

            logger.handlers.clear()

            logger.addHandler(handler)
            with ExitStack(lambda: logger.removeHandler(handler)):
                # Create the commands
                commands: list[Command] = [EchoOff()]

                for tool_info in tool_infos:
                    commands.append(Augment("PATH", str(tool_info.binary_directory)))

                    # Search for and apply environment files
                    env_config: dict[str, str] = {}

                    for potential_env_file in tool_info.GeneratePotentialEnvFiles():
                        if not potential_env_file.is_file():
                            continue

                        logger_sink.seek(0)
                        logger_sink.truncate(0)

                        # Load the content
                        try:
                            with potential_env_file.open("r", encoding="utf-8") as f:
                                this_env_config = dotenv_values(stream=f)

                            logger_content = logger_sink.getvalue()
                            if logger_content:
                                raise Exception(logger_content.rstrip())  # noqa: TRY301

                        except Exception as ex:
                            shell_dm.WriteError(
                                f"Unable to process the environment file '{potential_env_file}': {ex}\n"
                            )
                            continue

                        # Populate the placeholders
                        for key, value in this_env_config.items():
                            assert value is not None, key

                            this_env_config[key] = relative_path_re.sub(
                                lambda m, potential_env_file=potential_env_file: RelativePathReplacement(
                                    m,
                                    potential_env_file,
                                ),
                                value,
                            )

                        # Commit the results
                        env_config.update(cast(dict[str, str], this_env_config))

                    if env_config:
                        for key, value in env_config.items():
                            commands.append(Set(key, value))

                        commands.append(Raw("\n"))

                return commands



# noqa: D100
import shlex
import textwrap
import uuid

from dbrownell_Common.Types import override

from dbrownell_ToolsDirectory.Shell.Commands import (
    Message,
    Call,
    Execute,
    Set,
    Augment,
    Exit,
    ExitOnError,
    EchoOff,
    PersistError,
    PushDirectory,
    PopDirectory,
    Raw,
)
from dbrownell_ToolsDirectory.Shell.CommandVisitor import CommandVisitor


# ----------------------------------------------------------------------
class BatchCommandVisitor(CommandVisitor):
    """Command visitor for Windows batch scripts."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__()

        self._message_substitution_lookup: dict[str, str] = {
            "%": "%%",
            "&": "^&",
            "<": "^<",
            ">": "^>",
            "|": "^|",
            ",": "^,",
            ";": "^;",
            "(": "^(",
            ")": "^)",
            "[": "^[",
            "]": "^]",
        }

    # ----------------------------------------------------------------------
    @override
    def OnMessage(  # noqa: D102
        self,
        command: Message,
    ) -> str | None:
        output: list[str] = []

        for command_line in command.value.split("\n"):
            if not command_line.strip():
                # Note that the trailing white space seems to be necessary on some terminals
                output.append("echo. ")
                continue

            line = command_line.replace("^", "__caret_placeholder__")

            for source, dest in self._message_substitution_lookup.items():
                line = line.replace(source, dest)

            line = line.replace("__caret_placeholder__", "^")

            output.append(f"echo {line}")

        return " && ".join(output) + "\n"

    # ----------------------------------------------------------------------
    @override
    def OnCall(  # noqa: D102
        self,
        command: Call,
    ) -> str | None:
        result = f"call {command.command_line}\n"
        if command.exit_on_error:
            exit_on_error_result = self.Accept(ExitOnError())

            if exit_on_error_result:
                result += exit_on_error_result

        return result

    # ----------------------------------------------------------------------
    @override
    def OnExecute(  # noqa: D102
        self,
        command: Execute,
    ) -> str | None:
        commands: list[str] = shlex.split(command.command_line)

        if commands[0].lower().endswith((".bat", ".cmd")):
            result = f"cmd /c {command.command_line}"
        else:
            result = command.command_line

        result += "\n"

        if command.exit_on_error:
            exit_on_error_result = self.Accept(ExitOnError())

            if exit_on_error_result:
                result += exit_on_error_result

        return result

    # ----------------------------------------------------------------------
    @override
    def OnSet(  # noqa: D102
        self,
        command: Set,
    ) -> str | None:
        if command.value_or_values is None:
            return f"set {command.name}=\n"

        return f"set {command.name}={';'.join(command.EnumValues())}\n"

    # ----------------------------------------------------------------------
    @override
    def OnAugment(  # noqa: D102
        self,
        command: Augment,
    ) -> str | None:
        if command.append_values:
            add_statement_template = f"{{value}};%{command.name}%"
        else:
            add_statement_template = f"%{command.name}%;{{value}}"

        statement_template = textwrap.dedent(
            """\
            REM {{value}}
            echo ";%{name}%;" | findstr /C:";{{value}};" >nul
            if %ERRORLEVEL% == 0 goto skip_{{unique_id}}

            set {name}={add_statement_template}

            :skip_{{unique_id}}

            """,
        ).format(
            name=command.name,
            add_statement_template=add_statement_template,
        )

        statements: list[str] = [
            statement_template.format(
                value=value,
                unique_id=str(uuid.uuid4()).replace("-", ""),
            )
            for value in command.EnumValues()
        ]

        return "".join(statements)

    # ----------------------------------------------------------------------
    @override
    def OnExit(  # noqa: D102
        self,
        command: Exit,
    ) -> str | None:
        return textwrap.dedent(
            """\
            {success}
            {error}
            exit /B {return_code}
            """,
        ).format(
            success="if %ERRORLEVEL% EQ 0 (pause)" if command.pause_on_success else "",
            error="if %ERRORLEVEL% NEQ 0 (pause)" if command.pause_on_error else "",
            return_code=command.return_code or 0,
        )

    # ----------------------------------------------------------------------
    @override
    def OnExitOnError(  # noqa: D102
        self,
        command: ExitOnError,
    ) -> str | None:
        variable_name = command.variable_name or "ERRORLEVEL"

        return "if %{}% NEQ 0 (exit /B {})\n".format(
            variable_name,
            command.return_code if command.return_code is not None else f"%{variable_name}%",
        )

    # ----------------------------------------------------------------------
    @override
    def OnEchoOff(  # noqa: D102
        self,
        command: EchoOff,  # noqa: ARG002
    ) -> str | None:
        return "@echo off\n\n"

    # ----------------------------------------------------------------------
    @override
    def OnPersistError(  # noqa: D102
        self,
        command: PersistError,
    ) -> str | None:
        return f"set {command.variable_name}=%ERRORLEVEL%\n"

    # ----------------------------------------------------------------------
    @override
    def OnPushDirectory(  # noqa: D102
        self,
        command: PushDirectory,
    ) -> str | None:
        directory = command.value or "%~dp0"
        return f'pushd "{directory}"\n'

    # ----------------------------------------------------------------------
    @override
    def OnPopDirectory(  # noqa: D102
        self,
        command: PopDirectory,  # noqa: ARG002
    ) -> str | None:
        return "popd\n"

    # ----------------------------------------------------------------------
    @override
    def OnRaw(  # noqa: D102
        self,
        command: Raw,
    ) -> str | None:
        return command.value.removesuffix("\n") + "\n"
